
;; keromaa basic operation and switchboard

;; helpers

(define (bound? sym) (module-bound? (current-module) sym))
(define-macro (define-variable var default)
  `(if (not (bound? ',var)) (define ,var ,default)))

(define (call-stub name args) (debug-print 3 "Stub called: " name args))
(define-macro (stub name) `(define (,name . args) (call-stub ',name args)))

;; parsing

(stub make-talk-event)
(stub parse-action)

;; processing events

(stub handle-event)

;; talking

(stub report-offgame)
(stub report-event)

;; objects

(define-variable *object-pool* '())

(define (new-object class)
  (let ((obj (list class)))
    (set! *object-pool* (cons obj *object-pool*))
    obj))

(define (coproperty prop)
  (let* ((propname (symbol->string prop))
	 (co? (and (< 3 (string-length propname))
		   (char=? #\c (string-ref propname 0))
		   (char=? #\o (string-ref propname 1))
		   (char=? #\- (string-ref propname 2)))))
    (if co? (string->symbol (substring propname 3))
      (string->symbol (string-append "co-" propname)))))

(define (get-property prop obj) (or (assq-lookup prop (cdr obj)) '()))

(define (get-property-value prop obj)
  (let ((vals (get-property prop obj)))
    (if (null? vals) #f (car vals))))

(define (get-property-handle prop obj)
  (let ((handle (assq prop (cdr obj))))
    (if (not handle)
      (begin (set! handle (list prop '() #f))
	     (set-cdr! obj (cons handle (cdr obj)))))
    handle))

(define (add-property-internal! prop obj val cascade?)
  (let ((handle (get-property-handle prop obj)))
    (if (not (memq val (cadr handle)))
      (begin (set-car! (cdr handle) (cons val (cadr handle)))
	     (set-car! (cddr handle) (current-time))
	     (if (and cascade? (pair? val) (symbol? (car val)))
	       (add-property-internal! (coproperty prop) val obj #f))))))

(define (add-property! prop obj val)
  (add-property-internal! prop obj val #t))

(define (add-property-asymm! prop obj val)
  (add-property-internal! prop obj val #f))

(stub remove-property!)
(stub clear-property!)
(stub set-property!)

;; mapping IRC and keromaa identities

(define-variable *known-irc-users* (make-btree))
(define-variable *known-irc-chans* (make-btree))

(define (find-user-by-address irc-user context)
  (let* ((nick (irc-nick irc-user))
	 (host (string->symbol (irc-address-hostname irc-user)))
	 (chan (irc-channel-name context))
	 (user (btree-get *known-irc-users* nick #f)))
    (if (not user)
      (begin (set! user (generate-new-user irc-user chan))
	     (debug-print 5 "Automatically generated user:" nick)
	     (btree-set! *known-irc-users* nick user)))
    (let* ((hosts (get-property 'irc-hostname user))
	   (channels (get-property 'irc-channel user)))
      (if (and (not (memq chan channels))
	       (not (memq host hosts)))
	(report-offgame nick 'nameclash-problem)
	(begin (add-property! 'irc-hostname user host)
	       (add-property! 'irc-channel user chan)
	       (try-join-party! user (find-party-by-name chan))
	       user)))))

(define (find-party-by-name name)
  (let ((party (btree-get *known-irc-chans* name #f)))
    (if (not party)
      (begin (set! party (generate-new-party name))
	     (debug-print 5 "Automatically generated party:" name)
	     (btree-set! *known-irc-chans* name party)))
    party))

;; parties and members

(define (generate-new-party name)
  (let ((party (new-object 'party)))
    (add-property! 'name party name)
    party))

(define (generate-new-user irc-user chan)
  (let ((user (new-object 'user))
	(party (find-party-by-name chan)))
    (add-property! 'name user (irc-nick irc-user))
    (add-property! 'irc-hostname user
		   (string->symbol (irc-address-hostname irc-user)))
    (add-property! 'irc-channel user chan)
    (add-property! 'location user (party-coordinates party))
    (try-join-party! user party) ; order dep
    user))

(define (party-coordinates party)
  (let* ((members (get-property 'co-party party))
	 (member-locs (map (lambda (m) (get-property 'location m)) members))
	 (tidied-locs (filter (lambda (l) (not (null? l))) member-locs))
	 (member-loc (map car tidied-locs)))
    (if (null? member-loc) (random-location)
      (make-coord (average (map coord-x member-loc))
		  (average (map coord-y member-loc))
		  (average (map coord-z member-loc))))))

(define (try-join-party! user party)
  (if (near-party? user party)
    (add-property! 'party user party)
    (report-offgame user 'locationclash-problem)))

(define-variable *party-distance* 15)

(define (near-party? user party)
  (let ((party-loc (party-coordinates party))
	(user-loc (get-property-value 'location user)))
    (and party-loc user-loc
	 (>= *party-distance* (distance party-loc user-loc)))))

;; coordinates

(define (random-location) (make-coord 0 0 0))

(define (make-coord x y z) (vector x y z))
(define (coord-x coord) (vector-ref coord 0))
(define (coord-y coord) (vector-ref coord 1))
(define (coord-z coord) (vector-ref coord 2))

(define (distance co1 co2)
  (define (sq n) (* n n))
  (sqrt (+ (sq (- (coord-x co1) (coord-x co2)))
	   (sq (- (coord-y co1) (coord-y co2)))
	   (sq (- (coord-z co1) (coord-z co2))))))

(define (sum ls) (if (null? ls) 0 (+ (car ls) (sum (cdr ls)))))
(define (average ls) (/ (sum ls) (length ls)))

